#define G_LOG_DOMAIN "gpod-feed-item"

#include "config.h"

#include "gpod-feed-item.h"

#include <glib.h>
#include <glib-object.h>
#include <glib/gi18n.h>

/**
 * SECTION:gpod-feed-item
 * @title: GpodFeedItem
 * @short_description: A #GtkBox widget for list feeds
 * @include: gpod-feed-item.h
 * @see_also: #GpodFeed
 *
 * The #GpodFeedItem is an instance #GObject, and acts as a container for an
 * item found in the #GpodFeeds.
 **/

typedef struct _GpodFeedItemPrivate
{
    gchar *title;
    gchar *description;
    gchar *link;
    gchar *publish_date;
    gchar *enclosure;
    gchar *guid;
} GpodFeedItemPrivate;

struct _GpodFeedItem
{
    GObject parent;
};

G_DEFINE_TYPE_WITH_PRIVATE (GpodFeedItem, gpod_feed_item, G_TYPE_OBJECT)

enum
{
    PROP_0,
    PROP_TITLE,
    PROP_DESCRIPTION,
    PROP_LINK,
    PROP_PUBLISH_DATE,
    PROP_ENCLOSURE,
    PROP_GUID,
    NUM_PROPERTIES
};

static GParamSpec *properties[NUM_PROPERTIES] = { NULL, };

static void
gpod_feed_item_set_property (GObject      *object,
                             guint         property_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
    GpodFeedItem *self = GPOD_FEED_ITEM (object);
    GpodFeedItemPrivate *priv = gpod_feed_item_get_instance_private (self);

    switch (property_id)
    {
        case PROP_TITLE:
            priv->title = g_value_dup_string (value);
            break;

        case PROP_DESCRIPTION:
            priv->description = g_value_dup_string (value);
            break;

        case PROP_LINK:
            priv->link = g_value_dup_string (value);
            break;

        case PROP_PUBLISH_DATE:
            priv->publish_date = g_value_dup_string (value);
            break;

        case PROP_ENCLOSURE:
            priv->enclosure = g_value_dup_string (value);
            break;

        case PROP_GUID:
            priv->guid = g_value_dup_string (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
gpod_feed_item_get_property (GObject      *object,
                             guint         property_id,
                             GValue       *value,
                             GParamSpec   *pspec)
{
    GpodFeedItem *self = GPOD_FEED_ITEM (object);
    GpodFeedItemPrivate *priv = gpod_feed_item_get_instance_private (self);

    switch (property_id)
    {
        case PROP_TITLE:
            g_value_set_string (value, priv->title);
            break;

        case PROP_DESCRIPTION:
            g_value_set_string (value, priv->description);
            break;

        case PROP_LINK:
            g_value_set_string (value, priv->link);
            break;

        case PROP_PUBLISH_DATE:
            g_value_set_string (value, priv->publish_date);
            break;

        case PROP_ENCLOSURE:
            g_value_set_string (value, priv->enclosure);
            break;

        case PROP_GUID:
            g_value_set_string (value, priv->guid);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

/*
 * gpod_feed_item_finalize:
 * @object: An instance of #GpodFeedItem
 *
 * Does any last minute clean-ups when the class ends.
 */
static void
gpod_feed_item_finalize (GObject *object)
{
    G_OBJECT_CLASS (gpod_feed_item_parent_class)->finalize(object);
}

/*
 * gpod_feed_item_init:
 * @self: An instance of #GpodFeedItem
 *
 * Initializes various aspects of the #GpodFeedItem. Anything the #GpodFeedItem
 * would need before becoming presented should be done here.
 */
static void
gpod_feed_item_init (GpodFeedItem *self)
{

}

/*
 * gpod_feed_item_class_init:
 * @klass: An instance of #GpodFeedItemClass
 *
 * Initializes the various virtural functions in #GpodFeedItemClass. In
 * addtition it connects the various #GtkWidgets from the gresource file using
 * gtk_widget_class_set_template_from_resource().
 */

static void
gpod_feed_item_class_init (GpodFeedItemClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->set_property = gpod_feed_item_set_property;
    object_class->get_property = gpod_feed_item_get_property;
    object_class->finalize = gpod_feed_item_finalize;

    properties[PROP_TITLE] =
        g_param_spec_string ("title",
                             "Title",
                             "FeedItem title.",
                             NULL,
                             G_PARAM_READWRITE);

    properties[PROP_DESCRIPTION] =
        g_param_spec_string ("description",
                             "Description",
                             "Description of the feed.",
                             NULL,
                             G_PARAM_READWRITE);

    properties[PROP_LINK] =
        g_param_spec_string ("link",
                             "Link",
                             "Link to the feed site.",
                             NULL,
                             G_PARAM_READWRITE);

    properties[PROP_PUBLISH_DATE] =
        g_param_spec_string ("publish-date",
                             "Publish Date",
                             "Publish date of the feed.",
                             NULL,
                             G_PARAM_READWRITE);

    properties[PROP_ENCLOSURE] =
        g_param_spec_string ("last-build-date",
                             "Last Build Date",
                             "Date the feed was last built.",
                             NULL,
                             G_PARAM_READWRITE);

    properties[PROP_GUID] =
        g_param_spec_string ("image",
                             "Image",
                             "FeedItem image.",
                             NULL,
                             G_PARAM_READWRITE);

    g_object_class_install_properties (object_class,
                                       NUM_PROPERTIES,
                                       properties);
}

/**
 * gpod_feed_item_new:
 *
 * Creates a new #GpodFeedItem, a custom widget to show all feeds user is
 * subscribed to.
 *
 * ## Usage
 * |[
 * GpodFeedItem *feed;
 *
 * feed = gpod_feed_item_new ();
 * ]|
 *
 * Returns: #GpodFeedItem
 */
GpodFeedItem *
gpod_feed_item_new ()
{
    return g_object_new (GPOD_TYPE_FEED_ITEM,
                         NULL);
}
