#define G_LOG_DOMAIN "gpod-feed-parser"

#include "config.h"

#include "gpod-feed.h"
#include "gpod-feed-item.h"
#include "gpod-feed-parser.h"

#include <libsoup/soup.h>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include <gio/gio.h>
#include <glib.h>
#include <glib/gstdio.h>

#define BUFFER_SIZE 1024

/**
 * SECTION:gpod-feed-parser
 * @title: GpodFeedParser
 * @short_description: A #GtkBox widget for list feeds
 * @include: gpod-feed-parser.h
 * @see_also: #GpodFeed, #GpodFeedWidget
 *
 * The #GpodFeedParser is an instance #GtkBox, and acts as the primary
 * window/widget for listing feeds in the #GpodWindow<!-- -->'s #GtkStack.
 **/

struct _GpodFeedParser
{
    GObject parent;
};

G_DEFINE_TYPE (GpodFeedParser, gpod_feed_parser, G_TYPE_OBJECT)

typedef enum
{
    GPOD_ERROR_PARSE_FILE_FAILED,
    GPOD_ERROR_PARSE_STREAM_FAILED,
    GPOD_ERROR_DOCUMENT_EMPTY,
    GPOD_ERROR_DOCUMENT_NOT_RSS
} GpodFeedParserError;


typedef gint (*GpodChunkCallback) (gchar     *buffer,
                                   gpointer   user_data);

gint
file_read_cb (gchar    *buffer,
              gpointer  user_data)
{
    return fread (buffer, 1, sizeof (buffer), (FILE *) user_data);
}

gint
stream_read_cb (gchar    *buffer,
                gpointer  user_data)
{
    return g_input_stream_read ((GInputStream *) user_data,
                                buffer,
                                sizeof (buffer),
                                NULL,
                                NULL);
}

/*
 * parse_xml_chunks()
 * @self: A #GpodFeedParser
 * @chunk_cb: A #GpodChunkCallback
 * @user_data: A gpointer to the ouput stream, e.g. (FILE *), (GInputStream *)
 *
 * Reads a file and returns an xml document.
 *
 * Returns: (xmlDoc *): The xml document
 */
xmlDoc *
parse_xml_chunks (GpodFeedParser     *self,
                  GpodChunkCallback   chunk_cb,
                  gpointer            user_data)
{
    g_assert (GPOD_IS_FEED_PARSER (self));

    xmlParserCtxt *ctxt;
    gchar buffer[BUFFER_SIZE];
    gint bytes_read = 0;

    /* Create new xml parser */
    ctxt = xmlCreatePushParserCtxt (NULL, NULL, buffer, bytes_read, NULL);

    if (ctxt == NULL)
    {
        g_warning ("Failed to create parser context.");
        return NULL;
    }

    /* Read file ny chunks */
    while ((bytes_read = chunk_cb (buffer, user_data)) > 0)
    {
        xmlParseChunk (ctxt, buffer, bytes_read, 0);
        //TODO: Remove
        //gchar *out = g_strndup (buffer, bytes_read);
        //g_print ("%s", out);
    }

    xmlParseChunk (ctxt, buffer, 0, 1);

    if (!ctxt->wellFormed)
    {
        g_warning ("XML document malformed.");
        xmlFreeParserCtxt (ctxt);
        return NULL;
    }

    xmlFreeParserCtxt (ctxt);

    return ctxt->myDoc;
}

/*
 * parse_xml_from_file()
 * @self: A #GpodFeedParser
 * @file: The file to read from
 *
 * Reads a file and returns an xml document.
 *
 * Returns: (xmlDoc *): The xml document
 */
xmlDoc *
parse_xml_from_file (GpodFeedParser  *self,
                     gchar           *filename,
                     GError         **error)
{
    g_assert (GPOD_IS_FEED_PARSER (self));
    g_return_val_if_fail (filename != NULL, NULL);

    FILE *fp;
    xmlDoc *doc;

    /* Check file */
    if (!g_file_test (filename, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR))
    {
        g_warning ("Unable to open file to parse, either does not exists or not a file.");
        return NULL;
    }

    /* Open file to read */
    fp = g_fopen (filename, "r");

    if (fp == NULL)
    {
        g_warning ("Unable to open to parse.");
        return NULL;
    }

    doc = parse_xml_chunks (self, file_read_cb, fp);

    if (doc == NULL)
    {
        g_set_error_literal (error,
                             g_quark_from_string (G_LOG_DOMAIN),
                             GPOD_ERROR_PARSE_FILE_FAILED,
                             "Failed to parse stream.");
    }

    fclose (fp);

    return doc;
}

/*
 * parse_xml_from_stream()
 * @self: A #GpodFeedParser
 * @file: The stream to read from
 *
 * Reads a #GInputStream and returns an xml document.
 *
 * Returns: (xmlDoc *): The xml document
 */
xmlDoc *
parse_xml_from_stream (GpodFeedParser  *self,
                       GInputStream    *stream,
                       GError         **error)
{
    g_assert (GPOD_IS_FEED_PARSER (self));
    g_return_val_if_fail (G_IS_INPUT_STREAM (stream), NULL);
    g_return_val_if_fail (stream != NULL, NULL);

    xmlDoc *doc;

    doc = parse_xml_chunks (self, stream_read_cb, stream);

    if (doc == NULL)
    {
        g_set_error_literal (error,
                             g_quark_from_string (G_LOG_DOMAIN),
                             GPOD_ERROR_PARSE_STREAM_FAILED,
                             "Failed to parse stream.");
    }

    return doc;
}

/*
 * parse_xml_from_url()
 * @self: A #GpodFeedParser
 * @url: The url to fetch
 *
 * Fetches an XML document from @url
 *
 * Returns: (xmlDoc *): The xml document
 */
xmlDoc *
parse_xml_from_url (GpodFeedParser  *self,
                    gchar           *url)
{
    g_assert (GPOD_IS_FEED_PARSER (self));
    g_return_val_if_fail (url != NULL, NULL);

    g_debug ("Downloading feed from url \"%s\"", url);

    xmlDoc *doc;
    SoupSession *session;
    GError *error = NULL;
    SoupRequest *request;
    GInputStream *stream;

    session = soup_session_new ();

    request = soup_session_request (session, url, &error);

    if (error)
    {
        g_warning ("Failed to creating SOUP request from feed with url \"%s\" because: %s",
                   url, error->message);

        g_error_free (error);

        return NULL;
    }

    stream = soup_request_send (request, NULL, &error);

    if (error)
    {
        g_warning ("Failed to return SOUP request from feed with url \"%s\" because: %s",
                   url, error->message);

        g_error_free (error);

        return NULL;
    }

    doc = parse_xml_from_stream (self, stream, &error);

    if (error)
    {
        g_warning ("Failed to parse XML document from feed with url \"%s\" because: %s",
                   url, error->message);

        g_error_free (error);

        return NULL;
    }

    return doc;
}

static GpodFeed *
parse_xml (xmlDoc  *doc,
           GError **error)
{
    GpodFeed *feed;
    xmlNode *root;
    xmlNode *channel;

    feed = gpod_feed_new ();

    root = xmlDocGetRootElement (doc);

    if (root == NULL)
    {
        g_set_error_literal (error,
                             g_quark_from_string (G_LOG_DOMAIN),
                             GPOD_ERROR_DOCUMENT_EMPTY,
                             "Feed document is empty.");
        return NULL;
    }

    if (xmlStrcmp (root->name, (const xmlChar *)"rss"))
    {
        g_set_error_literal (error,
                             g_quark_from_string (G_LOG_DOMAIN),
                             GPOD_ERROR_DOCUMENT_NOT_RSS,
                             "Feed document is not a RSS document.");
        return NULL;
    }

    feed = gpod_feed_new ();

    channel = root->xmlChildrenNode;

    // List channels
    while (channel != NULL)
    {
        if (xmlStrcmp (channel->name, (const xmlChar *)"channel") == 0)
        {
            xmlNode *node = channel->xmlChildrenNode;
            gchar *title = NULL;
            gchar *description = NULL;
            gchar *link = NULL;
            gchar *publish_date = NULL;
            gchar *last_build_date = NULL;

            while (node != NULL)
            {
                if (xmlStrcmp(node->name, (const xmlChar *)"title") == 0)
                    title = g_strdup ((gchar *) xmlNodeGetContent (node));

                if (xmlStrcmp(node->name, (const xmlChar *)"description") == 0)
                    description = g_strdup ((gchar *) xmlNodeGetContent (node));

                if (xmlStrcmp(node->name, (const xmlChar *)"link") == 0)
                    link = g_strdup ((gchar *) xmlNodeGetContent (node));

                if (xmlStrcmp(node->name, (const xmlChar *)"pubDate") == 0)
                    publish_date = g_strdup ((gchar *) xmlNodeGetContent (node));

                if (xmlStrcmp(node->name, (const xmlChar *)"lastBuildDate") == 0)
                    last_build_date = g_strdup ((gchar *) xmlNodeGetContent (node));

                g_object_set (feed,
                              "title", title,
                              "description", description,
                              "link", link,
                              "publish_date", publish_date,
                              "last_build_date", last_build_date,
                              NULL);

                node = node->next;
            }
        }
        channel = channel->next;
    }


    return feed;
}

GpodFeed *
gpod_feed_parser_real_from_url (GpodFeedParser  *self,
                                gchar           *url)
{
    g_assert (GPOD_IS_FEED_PARSER (self));
    g_return_val_if_fail (url != NULL, NULL);

    g_debug ("Downloading feed from url \"%s\"", url);

    xmlDoc *doc;
    GpodFeed *feed;
    GError *error = NULL;

    doc = parse_xml_from_url (self, url);

    feed = parse_xml (doc, &error);

    if (error)
    {
        g_info ("Failed to parse XML document from feed with url \"%s\" because: %s",
                   url, error->message);

        g_error_free (error);

        xmlFreeDoc (doc);
        xmlCleanupParser ();
        xmlMemoryDump ();

        return NULL;
    }

    xmlFreeDoc (doc);
    xmlCleanupParser ();
    xmlMemoryDump ();

    return feed;
}

/*
 * gpod_feed_parser_finalize:
 * @object: An instance of #GpodFeedParser
 *
 * Does any last minute clean-ups when the class ends.
 */
static void
gpod_feed_parser_finalize (GObject *object)
{
    G_OBJECT_CLASS (gpod_feed_parser_parent_class)->finalize(object);
}

/*
 * gpod_feed_parser_init:
 * @self: An instance of #GpodFeedParser
 *
 * Initializes various aspects of the #GpodFeedParser. Anything the #GpodFeedParser
 * would need before becoming presented should be done here.
 */
static void
gpod_feed_parser_init (GpodFeedParser *self)
{

}

/*
 * gpod_feed_parser_class_init:
 * @klass: An instance of #GpodFeedParserClass
 *
 * Initializes the various virtural functions in #GpodFeedParserClass. In
 * addtition it connects the various #GtkWidgets from the gresource file using
 * gtk_widget_class_set_template_from_resource().
 */
static void
gpod_feed_parser_class_init (GpodFeedParserClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = gpod_feed_parser_finalize;
}

/**
 * gpod_feed_parser_new:
 *
 * Creates a new #GpodFeedParser, a custom widget to show all feeds user is
 * subscribed to.
 *
 * ## Usage
 * |[
 * GpodFeedParser *feed_parser;
 *
 * feed_parser = gpod_feed_parser_new ();
 * ]|
 *
 * Returns: #GpodFeedParser
 */
GpodFeedParser *
gpod_feed_parser_new ()
{
    return g_object_new (GPOD_TYPE_FEED_PARSER,
                         NULL);
}

GpodFeed *
gpod_feed_parser_from_url (GpodFeedParser  *self,
                           gchar           *url)
{
    return gpod_feed_parser_real_from_url (self, url);
}
