#define G_LOG_DOMAIN "gpod-window"

#include "config.h"

#include "gpod-feed.h"
#include "gpod-feed-item.h"
#include "gpod-feed-manager.h"

#include "gpod-library-view.h"
#include "gpod-feed-view.h"

#include "gpod-application.h"
#include "gpod-window.h"

#include <gtk/gtk.h>
#include <glib/gi18n.h>

/**
 * SECTION:gpod-window
 * @title: GpodWindow
 * @short_description: A #GtkApplicationWindow instance created for
 * #GpodApplication
 * @include: gpod-window.h
 * @see_also: #GtkApplicationWindow
 *
 * The #GpodApplication is an instance #GtkApplicationWindow, and acts as
 * the primary window for the current #GtkApplication.
 **/

typedef struct
{
    gint width;
    gint height;
    gint x;
    gint y;
    gboolean maximized;
    gboolean fullscreen;
} GpodWindowState;

typedef struct _GpodWindowPrivate
{
    GpodFeedManager *manager;

    GtkWidget *box;
    GtkWidget *headerbar;
    GtkWidget *overlay;
    GtkWidget *actionbar;

    GtkStack *stack;
    GtkWidget *player;

    GpodLibraryView *library_view;
    GpodFeedView *feed_view;

    guint timeout_id;
} GpodWindowPrivate;

struct _GpodWindow
{
    GtkApplicationWindow parent_instance;

    GpodWindowState state;
};

#define STATE_TIMEOUT 100 /* ms */

G_DEFINE_TYPE_WITH_PRIVATE (GpodWindow, gpod_window, GTK_TYPE_APPLICATION_WINDOW)

void
add_feeds (GpodWindow *self)
{
    GpodWindowPrivate* priv = gpod_window_get_instance_private (self);
    GpodFeed *feed;

    feed = g_object_new (GPOD_TYPE_FEED,
                         "title", "Example",
                         NULL);

    gpod_library_view_add_feed (priv->library_view, feed);

    feed = g_object_new (GPOD_TYPE_FEED,
                         "title", "Example 2",
                         NULL);

    gpod_library_view_add_feed (priv->library_view, feed);

    feed = g_object_new (GPOD_TYPE_FEED,
                         "title", "Example 3",
                         NULL);

    gpod_library_view_add_feed (priv->library_view, feed);

    feed = g_object_new (GPOD_TYPE_FEED,
                         "title", "Example 4",
                         NULL);

    gpod_library_view_add_feed (priv->library_view, feed);

    gtk_stack_set_visible_child_name (priv->stack, "library");
}

/*
 * on_close:
 * @action: A pointer to the object that originated the callback. In this
 * instance a #GSimpleAction, which will be added to the #GAction menu
 * @parameter: arg_data from the #GActionEntry
 * @user_data: User data set when the signal handler was connected, in this
 * instance it should be a #GtkApplication
 *
 * An action wrapper to close the current window.
 */
static void
on_close (GSimpleAction *action,
          GVariant      *state,
          gpointer       user_data)
{
    g_debug ("Requesting window be closed");
    gtk_window_close (GTK_WINDOW (user_data));
}

/*
 * on_fullscreen:
 * @action: A pointer to the object that originated the callback. In this
 * instance a #GSimpleAction, which will be added to the #GAction menu
 * @parameter: arg_data from the #GActionEntry
 * @user_data: User data set when the signal handler was connected, in this
 * instance it should be a #GtkApplication
 *
 * An action wrapper to toggle the window to fullscreen.
 */
static void
on_fullscreen (GSimpleAction *action,
               GVariant      *parameter,
               gpointer       user_data)
{
    gpod_window_toggle_fullscreen (GPOD_WINDOW (user_data));
}

const GActionEntry window_entries[] =
{
    { "close", on_close },
    { "fullscreen", on_fullscreen },
};

/*
 * gpod_window_set_fullscreen:
 * @self: An instance of #GpodWindow
 *
 * Sets the @self instance to fullscreen mode.
 */
void
gpod_window_toggle_fullscreen (GpodWindow *self)
{
    GPOD_IS_WINDOW (self);

    GdkWindow *window = gtk_widget_get_window (GTK_WIDGET (self));
    GdkWindowState state = gdk_window_get_state (window);

    if (state & GDK_WINDOW_STATE_FULLSCREEN) {
        gtk_window_unfullscreen (GTK_WINDOW (self));
        g_debug ("Window is requesting to unfullscreen");
    } else {
        gtk_window_fullscreen (GTK_WINDOW (self));
        g_debug ("Window is requesting to fullscreen");
    }

    /*
     * DEBUG CODE:
     * Check new state, after request
     */
    state = gdk_window_get_state (window);

    if (state & GDK_WINDOW_STATE_FULLSCREEN) {
        g_debug ("Window is unfullscreen");
    } else {
        g_debug ("Window is fullscreen");
    }
}

/*
 * gpod_window_state_store:
 * @self: An instance of #GpodWindow
 *
 * Saves the `width`, `height`, `x`, `y`, `maximized` and `fullscreen` values
 * from from the #GpodWindowState struct into the gsettings:
 * PACKAGE_ID ".window-state", eg. org.gnome.app.window-state
 */
static void
gpod_window_state_store (GpodWindow *self)
{
    GSettings *settings = g_settings_new (PACKAGE_ID ".window-state");
    gboolean status;

    status = g_settings_set_int (settings, "width", self->state.width);

    if (status)
        g_debug ("WindowSate: Saved width      = %i", self->state.width);
    else
        g_debug ("WindowState: Could not save width");


    status = g_settings_set_int (settings, "height", self->state.height);

    if (status)
        g_debug ("WindowSate: Saved height     = %i", self->state.height);
    else
        g_debug ("WindowState: Could not save height");

    status = g_settings_set_int (settings, "x", self->state.x);

    if (status)
        g_debug ("WindowSate: Saved position-x = %i", self->state.height);
    else
        g_debug ("WindowState: Could not save postion-x");

    status = g_settings_set_int (settings, "y", self->state.y);

    if (status)
        g_debug ("WindowSate: Saved position-y = %i", self->state.height);
    else
        g_debug ("WindowState: Could not save positon-y");

    status = g_settings_set_boolean (settings, "maximized", self->state.maximized);

    if (status)
        g_debug ("WindowSate: Saved maximized  = %s", (self->state.maximized) ? "TRUE" : "FALSE");
    else
        g_debug ("WindowState: Could not save maximized");

    status = g_settings_set_boolean (settings, "fullscreen", self->state.fullscreen);

    if (status)
        g_debug ("WindowSate: Saved fullscreen = %s", (self->state.fullscreen) ? "TRUE" : "FALSE");
    else
        g_debug ("WindowState: Could not save fullscreen");
}

/*
 * gpod_window_state_load:
 * @self: An instance of #GpodWindow
 *
 * Loads the `width`, `height`, `x`, `y`, `maximized` and `fullscreen` values
 * from gsettings into the #GpodWindowState struct, located in the
 * #GpodWindow struct.
 */
static void
gpod_window_state_load (GpodWindow *self)
{
    GSettings *settings = g_settings_new (PACKAGE_ID ".window-state");

    self->state.width = g_settings_get_int (settings, "width");
    g_debug ("WindowSate: Loaded width      = %i", self->state.width);

    self->state.height = g_settings_get_int (settings, "height");
    g_debug ("WindowSate: Loaded height     = %i", self->state.height);

    self->state.x = g_settings_get_int (settings, "x");
    g_debug ("WindowSate: Loaded position-x = %i", self->state.x);

    self->state.y = g_settings_get_int (settings, "y");
    g_debug ("WindowSate: Loaded position-y = %i", self->state.y);

    self->state.maximized = g_settings_get_boolean (settings, "maximized");
    g_debug ("WindowSate: Loaded maximized  = %s", (self->state.maximized) ? "TRUE" : "FALSE");

    self->state.fullscreen = g_settings_get_boolean (settings, "fullscreen");
    g_debug ("WindowSate: Loaded fullscreen = %s", (self->state.fullscreen) ? "TRUE" : "FALSE");
}

/*
 * gpod_window_state_set:
 * @user_data: Data passed into the function via g_timeout_add(), in this case
 * an instance of #GpodWindow
 *
 * Saves the information from the current window into the WindowState. Is called
 * from gpod_window_configure_event() and gpod_window_window_state_event().
 *
 * Returns: #gboolean; Sets %FALSE when the function does what it was going todo
 * (save window state) and set timeout in to 0.
 */
static gboolean
gpod_window_state_set (gpointer user_data)
{
    GpodWindow *self = GPOD_WINDOW (user_data);
    GpodWindowPrivate* priv = gpod_window_get_instance_private (self);

    GdkWindow *window = gtk_widget_get_window (GTK_WIDGET (self));
    GdkWindowState state = gdk_window_get_state (window);

    self->state.maximized  = (state & GDK_WINDOW_STATE_MAXIMIZED) ? 1 : 0;
    self->state.fullscreen = (state & GDK_WINDOW_STATE_FULLSCREEN) ? 1 : 0;

    if (self->state.maximized || self->state.fullscreen)
    {
        priv->timeout_id = 0;
        return FALSE;
    }

    gtk_window_get_size (GTK_WINDOW (self),
                         &self->state.width,
                         &self->state.height);

    // TODO: Make saving postion x and y work in Wayland?
    gtk_window_get_position (GTK_WINDOW (self),
                             &self->state.x,
                             &self->state.y);

    priv->timeout_id = 0;

    return FALSE;
}

/*
 * gpod_window_configure_event:
 * @widget: An instance of #GpodWindow
 * @event: The current state of the window as a #GdkEventWindowState
 *
 * Saves the window state/geometry when it changes. Calls
 * gpod_window_state_set().
 *
 * Returns: The #gboolean return value (retval); %TRUE the event happened or
 * %FALSE the event did not happen
 */
static gboolean
gpod_window_configure_event (GtkWidget         *widget,
                             GdkEventConfigure *event)
{
    GpodWindow *self = GPOD_WINDOW (widget);
    GpodWindowPrivate* priv = gpod_window_get_instance_private (self);
    gboolean retval;

    if (priv->timeout_id != 0)
    {
        g_source_remove (priv->timeout_id);
        priv->timeout_id = 0;
    }

    priv->timeout_id = g_timeout_add (STATE_TIMEOUT,
                                      gpod_window_state_set,
                                      self);

    retval = GTK_WIDGET_CLASS (gpod_window_parent_class)->configure_event (widget, event);

    return retval;
}

/*
 * gpod_window_destroy:
 * @widget: An instance of #GpodWindow
 * @event: The current state of the window as a #GdkEventWindowState
 *
 * Saves the window state/geometry when it changes, a.k.a if the window is
 * maximized or fullscreen. Calls gpod_window_state_set().
 *
 * Returns: The #gboolean return value (retval); %TRUE the event happened or
 * %FALSE the event did not happen
 */
static gboolean
gpod_window_window_state_event (GtkWidget           *widget,
                                GdkEventWindowState *event)
{
    GpodWindow *self = GPOD_WINDOW (widget);
    GpodWindowPrivate* priv = gpod_window_get_instance_private (self);
    gboolean retval;

    if (priv->timeout_id != 0)
    {
        g_source_remove (priv->timeout_id);
        priv->timeout_id = 0;
    }

    priv->timeout_id = g_timeout_add (STATE_TIMEOUT,
                                      gpod_window_state_set,
                                      self);

    retval = GTK_WIDGET_CLASS (gpod_window_parent_class)->window_state_event (widget, event);

    return retval;
}

/*
 * gpod_window_constructed:
 * @object: An instance of #GpodWindow
 *
 * The constructed virtual function sets first-run settings for the
 * #GpodWindow object.
 *
 * It sets:
 *  * Window dimensions
 *  * Window postion
 *  * Window maximized
 *  * Window fullscreen
 */
static void
gpod_window_constructed (GObject *object)
{
    GpodWindow *self = GPOD_WINDOW (object);

    self->state.width      = -1;
    self->state.height     = -1;
    self->state.x          =  0;
    self->state.y          =  0;
    self->state.maximized  =  FALSE;
    self->state.fullscreen =  FALSE;

    gpod_window_state_load (self);

    gtk_window_set_default_size (GTK_WINDOW (self),
                                 self->state.width,
                                 self->state.height);

    gtk_window_set_position (GTK_WINDOW (self), GTK_WIN_POS_MOUSE);

    if (self->state.x && self->state.y)
        gtk_window_move (GTK_WINDOW (self),
                         self->state.x,
                         self->state.y);

    if (self->state.maximized)
        gtk_window_maximize (GTK_WINDOW (self));

    if (self->state.fullscreen)
        gtk_window_fullscreen (GTK_WINDOW (self));

    G_OBJECT_CLASS (gpod_window_parent_class)->constructed (object);
}

/*
 * gpod_window_finalize:
 * @object: An instance of #GpodWindow
 *
 * Does any last minute clean-ups when the class ends.
 */
static void
gpod_window_finalize (GObject *object)
{
    GpodWindow *self = GPOD_WINDOW (object);
    GpodWindowPrivate* priv = gpod_window_get_instance_private (self);

    gpod_window_state_store (self);

    g_clear_object (&priv->headerbar);
    g_clear_object (&priv->player);

    G_OBJECT_CLASS (gpod_window_parent_class)->finalize(object);
}

/*
 * gpod_window_init:
 * @self: An instance of #GpodWindow
 *
 * Initializes various aspects of the window. Anything the window would need
 * before becoming presented should be done here.
 */
static void
gpod_window_init (GpodWindow *self)
{
    GApplication *application = g_application_get_default ();
    GpodWindowPrivate* priv = gpod_window_get_instance_private (self);
    GtkBuilder *builder = gtk_builder_new ();
    GError *error = NULL;

    const gchar *fullscreen_accels[] =
    {
        "F11",
        "<Primary>f",
        NULL
    };

    /* Init window.ui template */
    gtk_widget_init_template (GTK_WIDGET (self));

    /* Set window title */
    gtk_window_set_title (GTK_WINDOW (self), _("Podcast"));

    /* Add HeaderBar */
    if (!gtk_builder_add_from_resource (builder, PACKAGE_RESOURCES "/ui/headerbar.ui", &error))
    {
        g_warning ("%s", error->message);
        g_clear_error (&error);
    }
    else
    {
        priv->headerbar = GTK_WIDGET (g_object_ref (gtk_builder_get_object (builder, "headerbar")));
        gtk_window_set_titlebar (GTK_WINDOW (self), priv->headerbar);
    }

    /* Add Player */
    if (!gtk_builder_add_from_resource (builder, PACKAGE_RESOURCES "/ui/player.ui", &error))
    {
        g_warning ("%s", error->message);
        g_clear_error (&error);
    }
    else
    {
        priv->player = GTK_WIDGET (g_object_ref (gtk_builder_get_object (builder, "player")));
        gtk_container_add (GTK_CONTAINER (priv->actionbar), priv->player);
    }

    /* Add library view */
    priv->library_view = gpod_library_view_new ();

    gtk_stack_add_named (GTK_STACK (priv->stack),
                         GTK_WIDGET (priv->library_view),
                         "library");

    /* Add feed view */
    priv->feed_view = gpod_feed_view_new ();

    gtk_stack_add_named (GTK_STACK (priv->stack),
                         GTK_WIDGET (priv->feed_view),
                         "feed");


    /* Add episode view */

    /* Add empty states */
    // TODO: Disable by default and only enable if no content
    if (!gtk_builder_add_from_resource (builder, PACKAGE_RESOURCES "/ui/views/empty.ui", &error))
    {
        g_warning ("%s", error->message);
        g_clear_error (&error);
    }
    else
    {
        gtk_stack_add_named (GTK_STACK (priv->stack),
                             GTK_WIDGET (g_object_ref (gtk_builder_get_object (builder, "empty_view_feeds"))),
                             "empty_view_feeds");

        gtk_stack_add_named (GTK_STACK (priv->stack),
                             GTK_WIDGET (g_object_ref (gtk_builder_get_object (builder, "empty_view_search"))),
                             "empty_view_search");
    }

    /* Clean-up */
    g_clear_object (&builder);

    /* Add keybind */
    g_action_map_add_action_entries (G_ACTION_MAP (self),
                                     window_entries, G_N_ELEMENTS (window_entries),
                                     self);

    gpod_application_set_accelerator (GTK_APPLICATION (application), "win.close", "<Primary>w");
    gpod_application_set_accelerators (GTK_APPLICATION (application), "win.fullscreen", fullscreen_accels);

    /* Add Feeds */
    add_feeds (self);

    /* Clean-up */
    if (error != NULL)
        g_error_free (error);
}

/*
 * gpod_window_class_init:
 * @klass: An instance of #GpodWindowClass
 *
 * Initializes the various virtural functions in #GpodWindowClass. In
 * addtition it connects the various #GtkWidgets from the gresource file using
 * gtk_widget_class_set_template_from_resource().
 */
static void
gpod_window_class_init (GpodWindowClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

    object_class->constructed = gpod_window_constructed;
    object_class->finalize = gpod_window_finalize;

    widget_class->configure_event = gpod_window_configure_event;
    widget_class->window_state_event = gpod_window_window_state_event;

    gtk_widget_class_set_template_from_resource (widget_class,
                                                 PACKAGE_RESOURCES "/ui/window.ui");

    gtk_widget_class_bind_template_child_private (widget_class, GpodWindow, box);
    gtk_widget_class_bind_template_child_private (widget_class, GpodWindow, overlay);
    gtk_widget_class_bind_template_child_private (widget_class, GpodWindow, stack);
    gtk_widget_class_bind_template_child_private (widget_class, GpodWindow, actionbar);
}

/**
 * gpod_window_new:
 * @application: A #GtkApplication instance
 *
 * Creates a new #GpodWindow, a sub-object of #GtkApplicationWindow.
 *
 * ## Usage
 * |[
 * GpodWindow *window;
 *
 * window = gpod_window_new (application);
 * ]|
 *
 * When creating a gpod_window_new() instance it's recommended to set the
 * instance into a variable inside the #GpodApplication struct. The
 * window could also be added to a #GList, so multiple windows can belong to
 * the application.
 *
 * Returns: #GpodWindow; a sub-object of #GtkApplicationWindow
 */
GpodWindow *
gpod_window_new (GpodApplication *application)
{
    return g_object_new (GPOD_TYPE_WINDOW,
                         "application", application,
                         NULL);
}

