#define G_LOG_DOMAIN "gpod-feed"

#include "gpod-feed.h"

#include <glib.h>

/**
 * SECTION:gpod-feed
 * @short_description: a container to hold a RSS feed in GPOD
 *
 * #GpodFeed is the representation of a RSS feed. It contains a #JsonChannel,
 * which may contain #GpodItems.
 *
 * Since #GpodFeed<!-- -->'s can contain a #GpodChannel with many #GpodItems it
 * can be expensive. For this reason they are reference counted. You can control
 * the lifetime of a #GpodFeed using gpod_feed_ref() and gpod_feed_unref().
 */

G_DEFINE_BOXED_TYPE (GpodFeed, gpod_feed, gpod_feed_ref, gpod_feed_unref)

/**
 * gpod_feed_new: (constructor)
 *
 * Creates a new #GpodFeed, a container to hold a RSS feed in GPOD.
 *
 * Returns: (transfer full): the newly created #GpodFeed
 */
GpodFeed *
gpod_feed_new (void)
{
    GpodFeed *object;

    object = g_slice_new0 (GpodFeed);
    object->ref_count = 1;
    object->channel = NULL;

    return object;
}

/**
 * gpod_feed_ref:
 * @object: a #GpodFeed
 *
 * Increase by one the reference count of a #GpodFeed.
 *
 * Returns: (transfer none): the passed #GpodFeed, with the reference count
 *                           increased by one.
 */
GpodFeed *
gpod_feed_ref (GpodFeed *object)
{
    g_return_val_if_fail (object != NULL, NULL);
    g_return_val_if_fail (object->ref_count > 0, NULL);

    object->ref_count++;

    return object;
}


/**
 * gpod_feed_unref:
 * @object: a #GpodFeed
 *
 * Decreases by one the reference count of a #GpodFeed. If the
 * reference count reaches zero, the object is destroyed and all
 * its allocated resources are freed.
 */
void
gpod_feed_unref (GpodFeed *object)
{
    g_return_if_fail (object != NULL);
    g_return_if_fail (object->ref_count > 0);

    if (--object->ref_count == 0)
    {
        object->channel = NULL;

        g_slice_free (GpodFeed, object);
    }
}
